@@ -101,7 +101,7 @@ class LibvirtDriver(AbstractSystemsDriver):
 
     BOOT_LOADER_MAP = {
         'UEFI': {
-            'x86_64': '/usr/share/OVMF/OVMF_CODE.fd',
+            'x86_64': '/usr/share/OVMF/OVMF_CODE.secboot.fd',
             'aarch64': '/usr/share/AAVMF/AAVMF_CODE.fd'
         },
         'Legacy': {
@@ -111,6 +111,9 @@ class LibvirtDriver(AbstractSystemsDriver):
 
     }
 
+    SECURE_BOOT_ENABLED_NVRAM = '/usr/share/OVMF/OVMF_VARS.secboot.fd'
+    SECURE_BOOT_DISABLED_NVRAM = '/usr/share/OVMF/OVMF_VARS.fd'
+
     DEVICE_TYPE_MAP = {
         constants.DEVICE_TYPE_CD: 'cdrom',
         constants.DEVICE_TYPE_FLOPPY: 'floppy',
@@ -161,6 +164,12 @@ class LibvirtDriver(AbstractSystemsDriver):
             'SUSHY_EMULATOR_BOOT_LOADER_MAP', cls.BOOT_LOADER_MAP)
         cls.KNOWN_BOOT_LOADERS = set(y for x in cls.BOOT_LOADER_MAP.values()
                                      for y in x.values())
+        cls.SECURE_BOOT_ENABLED_NVRAM = cls._config.get(
+            'SUSHY_EMULATOR_SECURE_BOOT_ENABLED_NVRAM',
+            cls.SECURE_BOOT_ENABLED_NVRAM)
+        cls.SECURE_BOOT_DISABLED_NVRAM = cls._config.get(
+            'SUSHY_EMULATOR_SECURE_BOOT_DISABLED_NVRAM',
+            cls.SECURE_BOOT_DISABLED_NVRAM)
         cls.SUSHY_EMULATOR_IGNORE_BOOT_DEVICE = \
             cls._config.get('SUSHY_EMULATOR_IGNORE_BOOT_DEVICE', False)
         return cls
@@ -403,9 +412,11 @@ class LibvirtDriver(AbstractSystemsDriver):
         tree = ET.fromstring(self.get_xml_desc(domain))
 
         # Remove bootloader configuration
+        os_element_order = []
 
         for os_element in tree.findall('os'):
             for boot_element in os_element.findall('boot'):
+                os_element_order.append(boot_element.get('dev'))
                 os_element.remove(boot_element)
 
             if self.SUSHY_EMULATOR_IGNORE_BOOT_DEVICE:
@@ -428,20 +439,31 @@ class LibvirtDriver(AbstractSystemsDriver):
             raise error.FishyError(msg)
 
         target_device_elements = []
+        cur_hd_osboot_elements = []
+        cur_hd_order_elements = []
 
         # Remove per-disk boot configuration
+        # We should save at least hdd boot entries instead of just removing
+        # everything. In some scenarious PXE after provisioning stops replying
+        # and if there is no other boot device, then vm will fail to boot
+        # cdrom and floppy are ignored.
 
         for disk_element in devices_element.findall('disk'):
 
             device_attr = disk_element.get('device')
             if device_attr is None:
                 continue
+            boot_elements = disk_element.findall('boot')
 
             # NOTE(etingof): multiple devices of the same type not supported
             if device_attr == target:
                 target_device_elements.append(disk_element)
+            elif 'hd' in os_element_order:
+                cur_hd_osboot_elements.append(disk_element)
+            elif boot_elements:
+                cur_hd_order_elements.append(disk_element)
 
-            for boot_element in disk_element.findall('boot'):
+            for boot_element in boot_elements:
                 disk_element.remove(boot_element)
 
         target = self.INTERFACE_MAP.get(boot_source)
@@ -463,6 +485,15 @@ class LibvirtDriver(AbstractSystemsDriver):
 
             raise error.FishyError(msg)
 
+        # OS boot and per device boot order are mutually exclusive
+        if cur_hd_osboot_elements:
+            sorted_hd_elements = sorted(
+                cur_hd_osboot_elements,
+                key=lambda child: child.find('target').get('dev'))
+            target_device_elements.extend(sorted_hd_elements)
+        else:
+            target_device_elements.extend(cur_hd_order_elements)
+
         # NOTE(etingof): Make all chosen devices bootable (important for NICs)
 
         for order, target_device_element in enumerate(target_device_elements):
@@ -503,11 +534,34 @@ class LibvirtDriver(AbstractSystemsDriver):
 
         :raises: `error.FishyError` if boot mode can't be set
         """
-        domain = self._get_domain(identity, readonly=True)
 
-        # XML schema: https://libvirt.org/formatdomain.html#elementsOSBIOS
+        domain = self._get_domain(identity)
+
+        # XML schema:
+        # https://libvirt.org/formatdomain.html#operating-system-booting
         tree = ET.fromstring(self.get_xml_desc(domain))
+        self._build_os_element(identity, tree, boot_mode)
 
+        with libvirt_open(self._uri) as conn:
+
+            try:
+                conn.defineXML(ET.tostring(tree).decode('utf-8'))
+
+            except libvirt.libvirtError as e:
+                msg = ('Error changing boot mode at libvirt URI '
+                       '"%(uri)s": %(error)s' % {'uri': self._uri,
+                                                 'error': e})
+
+                raise error.FishyError(msg)
+
+    def _build_os_element(self, identity, tree, boot_mode, secure=None):
+        """Set the boot mode and secure boot on the os element
+
+        This also converts from the previous manual layout to the automatic
+        approach.
+
+        :raises: `error.FishyError` if boot mode can't be set
+        """
         try:
             loader_type = self.BOOT_MODE_MAP[boot_mode]
 
@@ -529,7 +583,6 @@ class LibvirtDriver(AbstractSystemsDriver):
         type_element = os_element.find('type')
         if type_element is None:
             os_arch = None
-
         else:
             os_arch = type_element.get('arch')
 
@@ -544,41 +597,83 @@ class LibvirtDriver(AbstractSystemsDriver):
                 boot_mode, os_arch)
             loader_path = None
 
-        loader_elements = os_element.findall('loader')
-        if len(loader_elements) > 1:
-            msg = ('Can\'t set boot mode because "loader" element must be '
-                   'present exactly once in domain "%(identity)s" '
+        # delete loader and nvram elements to rebuild from stratch
+        for element in os_element.findall('loader'):
+            os_element.remove(element)
+        for element in os_element.findall('nvram'):
+            os_element.remove(element)
+
+        loader_element = ET.SubElement(os_element, 'loader')
+        loader_element.set('type', loader_type)
+        if loader_path:
+            loader_element.text = loader_path
+            loader_element.set('readonly', 'yes')
+
+        if boot_mode == 'UEFI':
+            nvram_element = ET.SubElement(os_element, 'nvram')
+            if secure:
+                nvram_suffix = '.secboot.fd'
+                loader_element.set('secure', 'yes')
+                nvram_element.set('template', self.SECURE_BOOT_ENABLED_NVRAM)
+            else:
+                nvram_suffix = '.fd'
+                loader_element.set('secure', 'no')
+                nvram_element.set('template', self.SECURE_BOOT_DISABLED_NVRAM)
+
+            # force a different nvram path for secure vs not. This will ensure
+            # it gets regenerated from the template when secure boot mode
+            # changes
+            nvram_path = "/var/lib/libvirt/nvram-%s%s" % (identity,
+                                                          nvram_suffix)
+            nvram_element.text = nvram_path
+
+    def get_secure_boot(self, identity):
+        """Get computer system secure boot state for UEFI boot mode.
+
+        :returns: boolean of the current secure boot state
+
+        :raises: `FishyError` if the state can't be fetched
+        """
+        if self.get_boot_mode(identity) == 'Legacy':
+            msg = 'Legacy boot mode does not support secure boot'
+            raise error.NotSupportedError(msg)
+
+        domain = self._get_domain(identity, readonly=True)
+
+        # XML schema:
+        # https://libvirt.org/formatdomain.html#operating-system-booting
+        tree = ET.fromstring(domain.XMLDesc(libvirt.VIR_DOMAIN_XML_INACTIVE))
+
+        os_element = tree.find('os')
+
+        nvram = os_element.findall('nvram')
+        if len(nvram) > 1:
+            msg = ('Can\'t get secure boot state because "nvram" element '
+                   'must be present exactly once in domain "%(identity)s" '
                    'configuration' % {'identity': identity})
             raise error.FishyError(msg)
 
-        if loader_elements:
-            loader_element = loader_elements[0]
+        if not nvram:
+            return False
+        nvram_template = nvram[0].get('template')
+        return nvram_template == self.SECURE_BOOT_ENABLED_NVRAM
 
-            if loader_element.text not in self.KNOWN_BOOT_LOADERS:
-                msg = ('Unknown boot loader path "%(path)s" in domain '
-                       '"%(identity)s" configuration encountered while '
-                       'setting boot mode "%(mode)s", system architecture '
-                       '"%(arch)s". Consider adding this loader path to '
-                       'emulator config.' % {'identity': identity,
-                                             'mode': boot_mode,
-                                             'arch': os_arch,
-                                             'path': loader_element.text})
-                raise error.FishyError(msg)
+    def set_secure_boot(self, identity, secure):
+        """Set computer system secure boot state for UEFI boot mode.
 
-            if loader_path:
-                loader_element.set('type', loader_type)
-                loader_element.set('readonly', 'yes')
-                loader_element.text = loader_path
+        :param secure: boolean requesting the secure boot state
 
-            else:
-                # NOTE(etingof): path must be present or element must be absent
-                os_element.remove(loader_element)
+        :raises: `FishyError` if the can't be set
+        """
+        if self.get_boot_mode(identity) == 'Legacy':
+            msg = 'Legacy boot mode does not support secure boot'
+            raise error.NotSupportedError(msg)
 
-        elif loader_path:
-            loader_element = ET.SubElement(os_element, 'loader')
-            loader_element.set('type', loader_type)
-            loader_element.set('readonly', 'yes')
-            loader_element.text = loader_path
+        domain = self._get_domain(identity, readonly=True)
+
+        # XML schema: https://libvirt.org/formatdomain.html#elementsOSBIOS
+        tree = ET.fromstring(domain.XMLDesc(libvirt.VIR_DOMAIN_XML_INACTIVE))
+        self._build_os_element(identity, tree, 'UEFI', secure)
 
         with libvirt_open(self._uri) as conn:
 
@@ -586,7 +681,7 @@ class LibvirtDriver(AbstractSystemsDriver):
                 conn.defineXML(ET.tostring(tree).decode('utf-8'))
 
             except libvirt.libvirtError as e:
-                msg = ('Error changing boot mode at libvirt URI '
+                msg = ('Error changing secure boot at libvirt URI '
                        '"%(uri)s": %(error)s' % {'uri': self._uri,
                                                  'error': e})
 
@@ -701,7 +796,10 @@ class LibvirtDriver(AbstractSystemsDriver):
     def _process_bios(self, identity,
                       bios_attributes=DEFAULT_BIOS_ATTRIBUTES,
                       update_existing_attributes=False):
-        """Process Libvirt domain XML for BIOS attributes and update it if necessary
+        """Process Libvirt domain XML for BIOS attributes
+
+        Process Libvirt domain XML for BIOS attributes and update it if
+        necessary
 
         :param identity: libvirt domain name or ID
         :param bios_attributes: Full list of BIOS attributes to use if
@@ -712,6 +810,7 @@ class LibvirtDriver(AbstractSystemsDriver):
 
         :raises: `error.FishyError` if BIOS attributes cannot be saved
         """
+
         domain = self._get_domain(identity)
 
         result = self._process_bios_attributes(
